# Example PyPI (Python Package Index) Package

class Number(object):

    def __init__(self, n):
        self.value = n

    def val(self):
        return self.value

    def add(self, n2):
        self.value += n2.val()

    def __add__(self, n2):
        return self.__class__(self.value + n2.val())

    def __str__(self):
        return str(self.val())

    @classmethod
    def addall(cls, number_obj_iter):
        cls(sum(n.val() for n in number_obj_iter))



# -*- coding: utf-8 -*-
from operator import itemgetter
import ipaddress

def addlink(links, linkentry, owndomain):

    if linkentry['target'] == linkentry['source']:  # prevent circular links
        return links

    if linkentry['source'] == '/' or len(linkentry['source']) == 0:
        linkentry['source'] = owndomain

    if linkentry['target'] == '/' or len(linkentry['source']) == 0:
        linkentry['target'] = owndomain

    # check if links already exists:
    duplicate_found = False
    for li in links:
        if (li['source'] == linkentry['source']
            and li['target'] == linkentry['target']):
                duplicate_found = True
                li['value'] += linkentry['value']
                break
    
    # check for circular links:  ie. new target must not be in existing source
    for li in links:
        if li['source'] == linkentry['target']:
            duplicate_found = True
            break
    
    if duplicate_found == False:
        links.append(linkentry)
    #print("linkadd: " + str(duplicate_found) + " " + str(linkentry) )
    return links

def addnode(nodes, node, owndomain):

    if node['name'] == '/' or len(node['name']) == 0:
        node['name'] = owndomain

    if node['id'] == '/'  or len(node['id']) == 0:
        node['id'] = "".join(map(lambda char: char if char.isalnum()  else "", owndomain) ) # eliminate special chars

    # assert for sorting:
    if 'co' not in node:  
        node['co'] = 0     # count outgoing
        
    # check if node already exists:
    duplicate_found = False
    for no in nodes:
        if no['id'] == node['id']:
            duplicate_found = True
            break
    if duplicate_found == False:
        nodes.append(node)
    return nodes

# make nodes and links based on nav 20250221:
def makeNodesLinks2(dz, owndomain, omit, daysback=31):
    # temporary and for testing only:
    #dz['v0001'] = {}
    #dz['v0001']['days'] = {}
    #dz['v0001']['days']['20250228'] = {}
    #dz['v0001']['days']['20250228']['user'] = {}
    #dz['v0001']['days']['20250228']['user']['nav'] = [{"s": "pypi.org", "t": "/webstatsLF24.html", "p": "e", "c": 1}, {"s": "/", "t": "/kl/add/", "c": 1}, {"s": "/", "t": "/team/login/", "c": 1}, {"s": "/", "t": "/team/view/comparison%20team/", "c": 1}, {"s": "/", "t": "/kl/view/526f7142-abb1-4374-8020-e0c05266a988/", "c": 1}, {"s": "/xadm/log/a", "t": "/xadm/log/c", "c": 1}, {"s": "www.google.com", "t": "/", "c": 1}, {"s": "/", "t": "/kl/view/ef936637-2e83-482c-ac23-a3a7ffa10e17/", "c": 1}, {"s": "/", "t": "/kl/view/3deec972-5839-4a71-bd70-e2c886b83fda/", "c": 1}, {"s": "/", "t": "/kl/view/39575f54-e28f-4d97-bd6d-c7330cfd9d3c/", "c": 1}, {"s": "/", "t": "/kl/search/", "c": 1}, {"s": "/kl/search/", "t": "/", "c": 1}, {"s": "/team/viewtangledtree/ax/", "t": "/team/tag/ax/health/", "c": 1}, {"s": "/team/tag/ax/health/", "t": "/kl/view/bb0cd53a-cbf2-4fa2-9ae2-3a20fa7cc43f/", "c": 1}, {"s": "/kl/view/bb0cd53a-cbf2-4fa2-9ae2-3a20fa7cc43f/", "t": "/team/tag/ax/top10/", "c": 1}, {"s": "/team/tag/ax/top10/", "t": "/kl/view/f95c0c51-5f00-42f6-ae4c-50abb30ed665/", "c": 1}, {"s": "/", "t": "/kl/view/6c503f98-135e-4ade-81f0-9db92cc1fb05/", "c": 1}, {"s": "/xadm/sts", "t": "/xadm/log/a", "c": 2}, {"s": "/", "t": "/team/view/ax/", "c": 2}, {"s": "/", "t": "/kl/view/a05898ce-e766-4450-b0de-35a35c321097/", "c": 3}, {"s": "/team/view/ax/", "t": "/team/viewtangledtree/ax/", "c": 3}, {"s": "", "t": "/", "c": 56}, {"s": "google.com", "t": "/kl/view/a05898ce-e766-4450-b0de-35a35c321097/", "p": "e", "c": 5}, {"s": "google.de", "t": "/kl/view/6c503f98-135e-4ade-81f0-9db92cc1fb05/", "p": "e", "c": 1}, {"s": "cityfitness-eisenberg.de", "t": "/", "p": "e", "c": 1}, {"s": "google.com", "t": "/kl/view/3deec972-5839-4a71-bd70-e2c886b83fda/", "p": "e", "c": 5}, {"s": "google.com", "t": "/kl/view/39575f54-e28f-4d97-bd6d-c7330cfd9d3c/", "p": "e", "c": 1}, {"s": "google.com", "t": "/kl/view/abc909f2-10f6-4390-a153-ce597291a8db/", "p": "e", "c": 4}, {"s": "quora.com", "t": "/", "p": "e", "c": 1}, {"s": "yandex.ru", "t": "/kl/view/ef936637-2e83-482c-ac23-a3a7ffa10e17/", "p": "e", "c": 2}, {"s": "github.com", "t": "/webstatsLF24.html", "p": "e", "c": 1}, {"s": "google.de", "t": "/kl/view/abc909f2-10f6-4390-a153-ce597291a8db/", "p": "e", "c": 1}, {"s": "google.com", "t": "/", "p": "e", "c": 2}]
        
    links = []    # {'source': 'googlecom', 'target': 'teamviewax', 'c': 1}
    nodes = []    # {'id': 'googlecom', 'value':1, 'name': 'google.com'}

    days = 0
    for k, v in sorted(dz['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        # omit months or years:
        if len(k) <= 6:
            continue
        days += 1
        if 'nav' in dz['v0001']['days'][k]['user']:
            for e in dz['v0001']['days'][k]['user']['nav']:
                if is_valid_ip(e['s']) == True:  # to suppress ip; ip is not a domain anyway    
                    continue
                if '[' in e['s']:    # hack for IPv6 addresses
                    continue
                if ':' in e['s']:    # hack for IPv6 addresses/ports
                    continue

                tmpnode1 = {}
                tmpnode1['name'] = e['s']
                if tmpnode1['name'] == '/': 
                    tmpnode1['name'] = owndomain
                tmpnode1['id'] = "".join(map(lambda char: char if char.isalnum()  else "", tmpnode1['name']) ) # eliminate special chars
                tmpnode1['co'] = e['c']   # count outgoing
                nodes = addnode(nodes, tmpnode1, owndomain)

                tmpnode2 = {}
                tmpnode2['name'] = e['t']
                if tmpnode2['name'] == '/': 
                    tmpnode2['name'] = owndomain
                tmpnode2['id'] = "".join(map(lambda char: char if char.isalnum()  else "", tmpnode2['name']) ) # eliminate special chars
                tmpnode1['ci'] = e['c']  # count incoming
                nodes = addnode(nodes, tmpnode1, owndomain)
                
                tmplink = {}
                tmplink['source'] = tmpnode1['id']
                tmplink['target'] = tmpnode2['id']
                tmplink['value'] = e['c']
                links = addlink(links, tmplink, owndomain)
                    
        days += 1
        if days > daysback:
            break

    return nodes, links
    
# navigation chart as sankey diagram:
def makeNodesLinks(d, owndomain, omit, daysback=31):
    links = []    # {'source': 'googlecom', 'target': 'teamviewax', 'c': 1}
    nodes = []    # {'id': 'googlecom', 'value':1, 'name': 'google.com'}

    # externalFriendsHits:
    days = 0
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        # omit months or years:
        if len(k) <= 6:
            continue
        days += 1
        if 'externalFriendsHits' in d['v0001']['days'][k]['user']:
            for tk, tv in d['v0001']['days'][k]['user']['externalFriendsHits'].items():
                if is_valid_ip(tk) == True:  # to suppress ip; ip is not a domain anyway    
                    continue
                if '[' in tk:    # hack for IPv6 addresses
                    continue
                if ':' in tk:    # hack for IPv6 addresses/ports
                    continue

                tmpnode = {}
                tmpnode['name'] = tk
                if tmpnode['name'] == '/': 
                    tmpnode['name'] = owndomain
                tmpnode['id'] = "".join(map(lambda char: char if char.isalnum()  else "", tmpnode['name']) ) # eliminate special chars
                tmpnode['typ'] = 'root'
                nodes = addnode(nodes, tmpnode, owndomain)
                tmplink = {}
                tmplink['source'] =  "".join(map(lambda char: char if char.isalnum()  else "", tmpnode['name']) ) # eliminate special chars
                for tdk,tdv in tv['target'].items():
                    if any(oelm in tdk for oelm in omit):  # don not show parts of url 
                        continue
                    if tdk == '/':
                        tdk2 = owndomain
                    else:
                        tdk2 = tdk
                    tmplink['target'] =  "".join(map(lambda char: char if char.isalnum()  else "", tdk2) ) # eliminate special chars
                    tmplink['value'] = tdv
                    links = addlink(links, tmplink, owndomain)
                    
                    tmpnode = {}
                    tmpnode['name'] = tdk2
                    tmpnode['id'] = "".join(map(lambda char: char if char.isalnum()  else "", tdk2) ) # eliminate special chars
                    nodes = addnode(nodes, tmpnode, owndomain)
                days += 1
                if days > daysback:
                    break

        days = 0
        for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
            if len(k) < 8:   # not a day anymore
                break
            if 'user' in d['v0001']['days'][k] and 'navigation' in d['v0001']['days'][k]['user']:
                for sk,sv in d['v0001']['days'][k]['user']['navigation'].items():
                    n = sk.split('(())') 
                    if n[0] == n[1]:
                        continue
                    if len(n[0]) == 0 or len(n[1]) == 0:
                        continue
                    # allow only a-z and 0-9:
                    if n[0] == '/':    # to avoid empty na
                        n[0] = owndomain
                    if n[1] == '/':    # to avoid empty nb
                        n[1] = owndomain
                    if any(oelm in n[0] for oelm in omit):  # do not show parts of url 
                        continue
                    if any(oelm in n[1] for oelm in omit):  # do not show parts of url 
                        continue
                    tmplink = {}
                    tmplink['source'] = "".join(map(lambda char: char if char.isalnum()  else "", n[0]) ) # eliminate special chars
                    tmplink['target'] = "".join(map(lambda char: char if char.isalnum()  else "", n[1]) ) # eliminate special chars
                    tmplink['value'] = sv
                    links = addlink(links, tmplink, owndomain)

                    node = {}
                    node['id'] = "".join(map(lambda char: char if char.isalnum()  else "", n[0]) ) # eliminate special chars
                    node['name'] = n[0]
                    addnode(nodes, node, owndomain)

                    node = {}
                    node['id'] = "".join(map(lambda char: char if char.isalnum()  else "", n[1]) ) # eliminate special chars
                    node['name'] = n[1]
                    addnode(nodes, node, owndomain)
                
            days += 1
            if days > daysback:
                break

    return nodes, links

def is_valid_ip(address):
    try: 
        x = ipaddress.ip_address(address)
        return True
    except:
        return False



# -*- coding: utf-8 -*-
from operator import itemgetter
import ipaddress

# navigation chart in chord-dependency diagram:
# https://observablehq.com/@d3/chord-dependency-diagram/2

def navChord(d, owndomain, omit):
    data = []

    days = 0
    cnt_int = 0
    cnt_ext = 0
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        # omit months or years:
        if len(k) <= 6:
            continue
        days += 1
        if 'nav' in d['v0001']['days'][k]['user']:
            # external traffic only: top 30
            for e in sorted(d['v0001']['days'][k]['user']['nav'], key=itemgetter('c'), reverse=True):
                if 'p' not in e:
                    continue
                if e['c'] < 1:
                    continue
                if any(e['s'].startswith(om) for om in omit):  # omit parts of url
                    continue
                if any(e['t'].startswith(om) for om in omit):  # omit parts of url
                    continue

                if is_valid_ip(e['s']) == True:  # to suppress ip; ip is not a domain anyway    
                    continue
                if '[' in e['s']:    # hack for IPv6 addresses
                    continue
                if ':' in e['s']:    # hack for IPv6 addresses/ports
                    continue

                data,cnt_ext = addLinkChord(data, e, cnt_ext, owndomain)
                if cnt_ext > 30:  # top 30 entries
                    break

             # internal traffic only: top 30
            for e in sorted(d['v0001']['days'][k]['user']['nav'], key=itemgetter('c'), reverse=True):
                if 'p' in e:
                    continue
                if any(e['s'].startswith(om) for om in omit):  # omit parts of url
                    continue
                if any(e['t'].startswith(om) for om in omit):  # omit parts of url
                    continue

                data,cnt_int = addLinkChord(data, e, cnt_int, owndomain)
                if cnt_int > 30:  # top 30 entries
                    break
                    
        if days >= 30:
            break
        days += 1

    if len(data) == 0:
        return ''
   
    # d3js horizontal bubble char in case results are available
    h = "\n\n"
    h += '<div class="col-md-12 col-lg-12 col-xxl-12 pt-4">'
    h += '<h3>Daily User Navigation</h3>'
    h += '<p>Chord diagram showing user navigation for the last ' + str(days) + ' days on ' + owndomain + '. '
    h += 'Put the mouse pointer on the edge to get further details.</p>'
    h += '<div id="navchart-chord-container"></div>'
    h += '<script type="module">' + "\n"
    h += 'const data = ' + str(data) + ';' + "\n"
    h += 'function renderChart(data, options = {}) {'
    h += 'const rect = document.getElementById("navchart-chord-container").getBoundingClientRect();'
    h += 'const margins = { top: 20, right: 20, bottom: 40, left: 20 };'
    h += 'const width = Math.round(rect.width) - margins.left - margins.right;'
    h += 'const height = Math.round(width*0.9);'
    h += 'const innerRadius = Math.min(width, height) * 0.5 - 90;'
    h += 'const outerRadius = innerRadius + 10;' + "\n"

    # Dark mode configuration
    h += "const isDarkMode = () => window?.matchMedia?.('(prefers-color-scheme:dark)')?.matches ?? false;"
    h += 'const backgroundColor = isDarkMode ? "#1a1a1a" : "#ffffff";'
    h += 'const textColor = isDarkMode ? "#ffffff" : "#000000";'
    h += 'const fadedTextColor = isDarkMode ? "#cccccc" : "#666666";' + "\n"

    # Compute a dense matrix from the weighted links in data.
    h += 'const names = d3.sort(d3.union(data.map(d => d.source), data.map(d => d.target)));'
    h += 'const index = new Map(names.map((name, i) => [name, i]));'
    h += 'const matrix = Array.from(index, () => new Array(names.length).fill(0));'
    h += 'for (const {source, target, value} of data) matrix[index.get(source)][index.get(target)] += value;' + "\n"

    h += 'const chord = d3.chordDirected().padAngle(10 / innerRadius)'
    h += '.sortSubgroups(d3.descending).sortChords(d3.descending);' + "\n"

    h += 'const arc = d3.arc().innerRadius(innerRadius).outerRadius(outerRadius);'
    h += 'const ribbon = d3.ribbonArrow().radius(innerRadius - 1).padAngle(1 / innerRadius);'
    h += 'const colors = d3.quantize(d3.interpolateRainbow, names.length);' + "\n"

    h += 'const totalWidth = width + margins.left + margins.right;'
    h += 'const totalHeight = height;' + "\n"
    h += 'const svg = d3.select("#navchart-chord-container").append("svg")'
    h += '.attr("id","tt20250225")'
    h += '.attr("width", width)'
    h += '.attr("height", height)'
    h += '.attr("viewBox", [-width / 2, -height / 2, width, height])'
    h += '.attr("style", "width: 100%; height: auto; font: 10px sans-serif;");'  + "\n"

    # Add CSS styles for hover effects
    h += 'svg.append("style").text(`'
    h += '.chord-arc { opacity: 1;  }'
    h += '.chord-arc.fade { opacity: 0.2; }'
    h += '.chord-arc:hover { opacity: 0.9; }'
    h += '.chord-ribbon { opacity: 0.8; mix-blend-mode: ${isDarkMode ? "lighten" : "multiply"}; }'
    h += '.chord-ribbon.fade { opacity: 0.1; }'
    h += '.chord-ribbon.filtered { display: none; }'
    h += 'text { font-weight: bold; }'
    h += '.tick-text { fill: ${fadedTextColor}; font-weight: normal; }'
    h += '`);' + "\n"
    
    h += 'const chords = chord(matrix);'

    h += 'const group = svg.append("g").selectAll().data(chords.groups).join("g");' + "\n"
    h += 'group.append("path")'
    h += '.attr("class", "chord-arc")'
    h += '.attr("fill", d => colors[d.index])'
    h += '.attr("d", arc)'
    h += '.attr("id", d => `arc-${d.index}`)'
    h += '.on("mouseenter", function(event, d) {'
    # Fade all chords and arcs
    h += 'svg.selectAll(".chord-ribbon, .chord-arc").classed("fade", true);'
        
    # Build array of connected indices
    h += 'const connectedIndices = new Set();'
        
    # First identify all direct connections through ribbons
    h += 'svg.selectAll(`.chord-ribbon.source-${d.index}, .chord-ribbon.target-${d.index}`).each(function(ribbonData) {'
    # Only consider visible (not filtered) ribbons
    h += 'if (!d3.select(this).classed("filtered")) {'
    # Only add the connected index (not this arc's index)
    h += 'if (ribbonData.source.index === d.index) {'
    h += 'connectedIndices.add(ribbonData.target.index);'
    h += '} else {'
    h += 'connectedIndices.add(ribbonData.source.index);'
    h += '}'
    h += '}'
    h += '})'
    h += '.classed("fade", false);'
        
    # Add the current arc index
    h += 'connectedIndices.add(d.index);'
        
    # Unfade only directly connected arcs and current arc
    h += 'connectedIndices.forEach(index => {'
    h += 'svg.select(`#arc-${index}`).classed("fade", false);'
    h += '});'
    h += '})'  + "\n"
    h += '.on("mouseleave", function() {'
    # Reset all elements to normal state
    h += 'svg.selectAll(".chord-ribbon, .chord-arc")'
    h += '.classed("fade", false);'
    h += '});' + "\n"

  
    h += 'group.append("text")'
    h += '.each(d => (d.angle = (d.startAngle + d.endAngle) / 2))'
    h += '.attr("dy", "0.35em")'
    h += '.attr("transform", d => `rotate(${(d.angle * 180 / Math.PI - 90)})'
    h += 'translate(${outerRadius + 5}) ${d.angle > Math.PI ? "rotate(180)" : ""} `)'
    h += '.attr("text-anchor", d => d.angle > Math.PI ? "end" : null)'
    h += '.attr("fill", "black")'
    h += '.text(d => names[d.index]);'  + "\n"

    h += 'group.append("title")'
    h += '.text(d => `${names[d.index]}'
    h += 'outgoing: ${d3.sum(chords, c => (c.source.index === d.index) * c.source.value)} '
    h += 'incoming: ${d3.sum(chords, c => (c.target.index === d.index) * c.source.value)}`);' + "\n"

    h += 'const ribbons = svg.append("g")'
    h += '.attr("fill-opacity", 0.75)'
    h += '.selectAll()'
    h += '.data(chords)'
    h += '.join("path")'
    h += '.style("mix-blend-mode", "multiply")'
    h += '.attr("class", d => `chord-ribbon source-${d.source.index} target-${d.target.index}`)'
    h += '.attr("fill", d => colors[d.target.index])'
    h += '.attr("d", ribbon);' + "\n"

    h += 'ribbons.append("title")'
    h += '.text(function(d) {'
    # Get the names of source and target
    h += 'let sourceName = names[d.source.index];'
    h += 'let targetName = names[d.target.index];'
      
    # If the ribbon is filtered (hidden), don't show any tooltip
    h += 'if (d3.select(this).classed("filtered")) { return ""; }'
    
    h += 'let lineTitle = `${d.source.value} ${sourceName} → ${targetName}`;'
    h += 'if (d.source.index !== d.target.index) {'
    h += 'lineTitle += `${d.target.value} ${targetName} → ${sourceName}`;'
    h += '}'
    h += 'return lineTitle;'
    h += '});' + "\n"

    h += 'ribbons.on("mouseenter", function(event, d) {'
    # Fade all chords and arcs
    h += 'svg.selectAll(".chord-ribbon, .chord-arc").classed("fade", true);'
          
    # Keep only this chord visible
    h += 'd3.select(this).classed("fade", false);'
          
    # Keep only connected arcs visible
    h += 'const sourceArc = svg.select(`#arc-${d.source.index}`);'
    h += 'const targetArc = svg.select(`#arc-${d.target.index}`);'
    h += 'sourceArc.classed("fade", false);'
    h += 'targetArc.classed("fade", false);'
    h += '}).on("mouseleave", function() {'
    # Reset all elements to normal state
    h += 'svg.selectAll(".chord-ribbon, .chord-arc").classed("fade", false);'
    h += '});'  + "\n"
    h += '}' + "\n"   # end of function renderChart


    h += 'renderChart(data, { backgroundColor: "#f8f8f8" });' + "\n"
    h += "</script>"
    h += '</div>' + "\n"
    return h
    
def addLinkChord(data, e, cnt, owndomain):
    tmplink = {}
    tmplink['source'] = e['s']
    if tmplink['source'] == '/' or tmplink['source'] == None or tmplink['source'] == '':
        tmplink['source'] = owndomain
                    
    tmplink['target'] = e['t']
    if tmplink['target'] == '/' or tmplink['target'] == None or tmplink['target'] == '':
        tmplink['target'] = owndomain
                    
    tmplink['value'] = e['c']
    if tmplink['source'] == tmplink['target']:
       return data, cnt
                
    duplicate_found = False
    for li in data:
        if (li['source'] == tmplink['source']
            and li['target'] == tmplink['target']):
            duplicate_found = True
            li['value'] += tmplink['value']
            break
    if duplicate_found == False:
        data.append(tmplink)
        cnt += 1
    return data, cnt

def is_valid_ip(address):
    try: 
        x = ipaddress.ip_address(address)
        return True
    except:
        return False



# -*- coding: utf-8 -*-
from operator import itemgetter
from .validip import is_valid_ip

def dailyTopDomains(d, owndomain, omit):
    day_usr_i = 0
    tsource = {}
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        # consider only days:
        if len(k) < 8:
            continue
        if any(oelm in k for oelm in omit):  # don not show parts of url 
            continue
        # show at max 31 days:
        if day_usr_i >= 31:
            break
        day_usr_i += 1

        # top 10 source domains:
        if 'nav' in d['v0001']['days'][k]['user']:
            # sort by count e['c'] desc:
            for e in sorted(d['v0001']['days'][k]['user']['nav'], key=lambda x: x['c'], reverse=True):
                if 'p' not in e:   # internal links
                    continue
                if e['s'] in omit:
                    continue
                if e['t'] in omit:
                    continue
                if e['s'] not in tsource:
                    tsource[e['s']] = 0
                tsource[e['s']] += e['c']
    
    if len(tsource) == 0:
        return ''
    
    # Top 10 Domains on daily basis
    # row needs to be decleared in the calling function h+= '<div class="row pt-3">'
    h = "\n" + '<div class="col-md-12 col-lg-6 col-xxl-6">'
    h += '<h3>Top 10 Domains for the last ' + str(day_usr_i) + ' days</h3>'
    h += '<p>User hits refering to external domain:</p>'
    h += '<table class="table">'
    h += '<thead><tr><th scope="col" style="text-align: left">Rank</th><th scope="col" style="text-align: left">Domain</th><th scope="col" style="text-align: right">Hits</th></tr></thead>'
    i = 1
    for k, v in sorted(tsource.items(), key=itemgetter(1), reverse=True):
        if owndomain in k.lower():
            continue
        # prevent IP or domains with special characters:
        if any( x in k for x in {'[', ']', ':'} ):
            continue
        if is_valid_ip(k) == True:  # to suppress ip; ip is not a domain anyway    
            continue
        h += '<tr><td>' + str(i) + '.</td><td>' + str(k) + '</td><td style="text-align: right">' + str(format(v, ',')) + '</td></tr>'
        i += 1
        if i > 10:
            break
    h += '</table>'
    h += '</div>'  + "\n"   # end of col and card
 
    return h



# -*- coding: utf-8 -*-
from operator import itemgetter
import ipaddress

# navigation chart in tangled tree visualization:
def navchartTangledtree(nodes, links, owndomain, omit):

    # const levels = [[{'id': 'start'}], [{'id': 'business', 'parents': ['start']}], [{'id': 'comparison', 'parents': ['start']}]
    levels = [[{'id': 'start', 'name': 'start'}]]
    print("links: " + str(links))

    # append root nodes to levels L1:
    # root nodes are sources that never appear in target
    tmplevel1 = []
    for n1 in nodes:
        found = False
        for l1 in links:
            if l1['target'] == n1['id']:
                found = True
                break
        if found == False:
            tmp = {}
            tmp['id'] = n1['id']
            tmp['name'] = n1['name']
            tmp['parents'] = ['start']
            tmplevel1.append(tmp)
    levels.append(tmplevel1)

    print("levels1: " + str(levels) )

    # 2nd Level:
    tmplevel2 = []
    for l1 in tmplevel1:
        print("l1: " + str(l1))
        found = False
        for li2 in links:
            if li2['source'] == l1['id']:
                found = True
                break
        if found == True:
            # check if l1 is already there --> add to parents
            parent_found = False
            for l2a in tmplevel2:
                if l2a['id'] == li2['target']:
                    parent_found = True
                    l2a['parents'].append(l1['id'])
                    break
            if parent_found == False:
                tmp = {}
                tmp['id'] = li2['target']
                tmp['name'] = li2['target']
                tmp['parents'] = [l1['id']]
                for n2 in nodes:
                    if li2['target'] == n2['id']:
                        tmp['name'] = n2['name']
                        break
                tmplevel2.append(tmp)
            
    levels.append(tmplevel2)

    # 3rd Level:
    tmplevel3 = []
    for l2 in tmplevel2:
        print("l2: " + str(l2))
        found = False
        for li3 in links:
            if li3['source'] == l2['id']:
                found = True
                break
        if found == True:
            # check if l2 is already there --> add to parents
            parent_found = False
            for l3a in tmplevel3:
                if l3a['id'] == li3['target']:
                    parent_found = True
                    l3a['parents'].append(l2['id'])
                    break
            if parent_found == False:
                tmp = {}
                tmp['id'] = li3['target']
                tmp['name'] = li3['target']
                tmp['parents'] = [l2['id']]
                for n3 in nodes:
                    if li3['target'] == n3['id']:
                        tmp['name'] = n3['name']
                        break
                tmplevel3.append(tmp)
            
    levels.append(tmplevel3)
 

    # d3js horizontal bubble char in case results are available
    h = "\n\n"
    h += '<div class="col-md-12 col-lg-12 col-xxl-12">'
    h += '<div class="card mt-2"><div class="card-body">'
    h += '<h3 class="card-title">Navigation Chart</h3>'
    h += '<p class="card-text">User Navigation Chart for ' + owndomain + ':</p>'
    h += '<div id="navchart-tangledtree-container"></div>'
    h += '<script type="module">' + "\n"
    h += 'const levels = ' + str(levels) + ';' + "\n"
    h += 'const color = d3.scaleOrdinal(d3.schemeDark2);'

    h += 'const rect = document.getElementById("navchart-tangledtree-container").getBoundingClientRect();'
    h += 'const margins = { top: 20, right: 20, bottom: 40, left: 100 };'
    h += 'const width = Math.round(rect.width) - margins.left - margins.right;'
    h += 'const height = width;'  + "\n"   # make height at least as width
    h += 'const totalWidth = width + margins.left + margins.right;'
    h += 'const totalHeight = height + margins.top + margins.bottom;'
    h += 'const svg = d3.select("#navchart-tangledtree-container")'
    h += '.append("svg")'
    h += '.attr("width", totalWidth)'
    h += '.attr("height", totalHeight)'
    h += '.attr("viewBox", [0,0 , totalWidth, totalHeight])' + "\n"
    h += '.attr("style", "max-width: 100%; height: auto; font: 10px sans-serif; height: intrinsic;");' + "\n"

    h += 'const graphGroup = svg.append("g");'
    #h += '.attr("transform", "translate(20," + margins.top + ")");'

    h += 'levels.unshift([]);'

    # We add one pseudo node to every level to deal with parentless nodes
    h += 'levels.forEach((l, i) => {'
    h += 'l.forEach((n, j) => {'
    h += 'n.level = i;'
    h += 'if (n.parents !== undefined) {'
    h += '  n.parent = n.parents[0];'
    h += '} else {'
    h += '  n.parent = `pseudo-${i - 1}`;'
    h += '}'
    h += '});'
    h += 'l.unshift({'
    h += 'id: `pseudo-${i}`,'
    h += 'parent: i > 0 ? `pseudo-${i - 1}` : "",'
    h += 'level: i'
    h += '});'
    h += '});' + "\n"

    h += 'const nodes = levels.flat();'
    h += 'const colours = d3.scaleOrdinal()'
    h += '.domain(nodes.filter(n => n.parents)'
    h += '.map(n => n.parents.sort()'
    h += '.join("-")))'
    h += '.range(d3.schemePaired);'

    h += 'function getLinks(nodes) {'
    h += 'return nodes'
    h += '.filter(n => n.data.parents !== undefined)'
    h += '.map(n => n.data.parents.map(p => ({ source: nodes.find(n => n.id === p), target: n  })))'
    h += '.flat();'
    h += '}'  + "\n"

    h += 'const offsetPerPartner = 3;'
    h += 'const drawNodePath = d => {'
    h += 'const radius = 5;'
    # // The number of partners determines the node height
    # // But when a node has only one partner,
    # // treat it the same as when it has zero
    h += 'const nPartners = (d.data.partners && d.data.partners.length > 1) ? d.data.partners.length : 0;'  + "\n"

    #// We want to centre each node
    h += 'const straightLineOffset = (nPartners * offsetPerPartner) / 2;'

    h += 'const context = d3.path();'
    h += 'context.moveTo(-radius, 0);'
    h += 'context.lineTo(-radius, -straightLineOffset);'
    h += 'context.arc(0, -straightLineOffset, radius, -Math.PI, 0);'
    h += 'context.lineTo(radius, straightLineOffset);'
    h += 'context.arc(0, straightLineOffset, radius, 0, Math.PI);'
    h += 'context.closePath();'

    h += 'return context + "";'
    h += '};' + "\n"

    h += 'const drawLinkCurve = (x0, y0, x1, y1, offset, radius) => {'
    h += 'const context = d3.path();'
    h += 'context.moveTo(x0, y0);'
    h += 'context.lineTo(x1 - 2 * radius - offset, y0);'  + "\n"

    #// If there is not enough space to draw two corners, reduce the corner radius
    h += 'if (Math.abs(y0 - y1) < 2 * radius) { radius = Math.abs(y0 - y1) / 2; }'  + "\n"

    h += 'if (y0 < y1) {'
    h += 'context.arcTo(x1 - offset - radius, y0, x1 - offset - radius, y0 + radius, radius);'
    h += 'context.lineTo(x1 - offset - radius, y1 - radius);'
    h += 'context.arcTo(x1 - offset - radius, y1, x1 - offset, y1, radius);'
    h += '} else if (y0 > y1) {'
    h += 'context.arcTo(x1 - offset - radius, y0, x1 - offset - radius, y0 - radius, radius);'
    h += 'context.lineTo(x1 - offset - radius, y1 + radius);'
    h += 'context.arcTo(x1 - offset - radius, y1, x1 - offset, y1, radius);'
    h += '}' + "\n"
    h += 'context.lineTo(x1, y1);'
    h += 'return context + "";'
    h += '};' + "\n"

    h += 'const partnershipsPerLevel = {};'
    h += 'const getPartnershipOffset = (parent, partner) => {'
    h += 'let partnershipId, level;'
    h += 'if(partner !== undefined) {'
    #// On every level, every relationship gets its own offset. If a relationship
    #// spans multiple levels, the furthest level is chosen
    h += 'level = Math.max(parent.depth, partner.level);'
    h += 'if(!partnershipsPerLevel[level]) { partnershipsPerLevel[level] = []; }'
    h += 'partnershipId = [parent.id, partner.id].sort().join("-");'
    h += '} else {'
    h += 'level = parent.depth;'
    h += 'if(!partnershipsPerLevel[level]) { partnershipsPerLevel[level] = []; }'
    h += 'partnershipId = parent.id;'
    h += '}'

    #// Assume that the partnership already has a slot assigned
    h += 'const partnershipOffset = partnershipsPerLevel[level].indexOf(partnershipId);'
    h += 'if(partnershipOffset === -1) {'
    #// Apparently not
    h += 'return partnershipsPerLevel[level].push(partnershipId) - 1;'
    h += '}'
    h += 'return partnershipOffset;'
    h += '}' + "\n"

    h += 'const lineRadius = 10;'
    h += 'const offsetStep = 5;'
    h += 'const linkFn = link => {'
    h += 'const thisParent = link.source;'
    h += 'const partnerId = link.target.data.parents.find(p => p !== thisParent.id);'
    h += 'const partners = thisParent.data.partners || [];' + "\n"

    #// Let the first link start with this negative offset
    #// But when a node has only one partner,
    #// treat it the same as when it has zero
    h += 'const startOffset = (partners.length > 1) ? -(partners.length * offsetPerPartner) / 2 : 0;'  + "\n"

    h += 'const partner = partners.find(p => p.id === partnerId);'

    # Chaos has no partner, nor Zeus with Athena
    h += 'const nthPartner = partner !== undefined'
    h += '? partners.indexOf(partner)'
    h += ': (partners || []).length;'
    h += 'const partnershipOffset = getPartnershipOffset(thisParent, partner);' + "\n"

    h += 'return drawLinkCurve('
    h += 'thisParent.y,'
    h += 'thisParent.x + startOffset + offsetPerPartner * nthPartner,'
    h += 'link.target.y,'
    h += 'link.target.x,'
    h += 'offsetStep * partnershipOffset,'
    h += 'lineRadius'
    h += ');'
    h += '};' + "\n"

    h += 'function draw(root) {'
    #// Now every node has had it's position set, we can draw them now
    h += 'const nodes = root.descendants().filter(n => !n.id.startsWith("pseudo-"));'
    h += 'const links = getLinks(nodes).filter(l => !l.source.id.startsWith("pseudo-"));'  + "\n"

    h += 'const link = graphGroup.selectAll(".link").data(links);'
    h += 'link.exit().remove();'
    h += 'link.enter()'
    h += '.append("path")'
    h += '.attr("class", "link")'
    h += '.style("fill", "none")'  # from stylesheet
    h += '.merge(link)'
    h += '.attr("stroke", d => colours(d.target.data.parents.sort().join("-")))'
    h += '.attr("d", linkFn);'  + "\n"

    h += 'const node = graphGroup.selectAll(".node").data(nodes);'
    h += 'node.exit().remove();'
    h += 'const newNode = node.enter().append("g")'
    h += '.on("click", click)'
    h += '.attr("class", "node")'
    h += '.append("a")'
    h += '.attr("xlink:href", d => d.data.name ? (d.data.name.startsWith("/") ? `https://'+ owndomain + '${d.data.name}` : `https://${d.data.name}`) : "#")'
    h += '.attr("target", "_blank")' # Opens the link in a new tab
    h += '.attr("title", d => `${d.name}`)'
    h += ';'  + "\n"

    h += 'newNode.append("path").attr("d", drawNodePath);'
    h += 'newNode.append("text")'
    h += '.attr("dy", -3)'
    h += '.attr("x", 6)'
    h += '.style("font-size", "10px")'
    h += '.style("font-family", "sans-serif,Open Sans");'  + "\n"

    h += 'newNode.merge(node)'
    h += '.attr("transform", d => `translate(${d.y},${d.x})`)'
    #h += '.attr("transform", d => `translate(${d.y - width / 2},${d.x})`)'
    h += '.selectAll("text")'
    #h += '.text(d => d.id + (d.cnt ? ` (${d.cnt})` : ""));' # add cnt if available
    h += '.text(d => d.data.name + (d.cnt ? ` (${d.cnt})` : ""));' # add cnt if available
    h += '}' + "\n"

    h += 'const root = d3.stratify().parentId(d => d.parent)(nodes);'  + "\n"

    #// Map the different sets of parents,
    #// assigning each parent an array of partners
    h += 'getLinks(root.descendants())'
    h += '.filter(l => l.target.data.parents)'
    h += '.forEach(l => {'
    h += 'const parentNames = l.target.data.parents;'
    h += 'if (parentNames.length > 1) {'
    h += 'const parentNodes = parentNames.map(p => nodes.find(n => n.id === p));'  + "\n"

    h += 'parentNodes.forEach(p => {'
    h += 'if (!p.partners) {'
    h += 'p.partners = [];'
    h += '}'
    h += 'parentNodes'
    h += '.filter(n => n !== p && !p.partners.includes(n))'
    h += '.forEach(n => { p.partners.push(n); });'
    h += '});'
    h += '}'
    h += '});'  + "\n"

    #// Take nodes with more partners first,
    #// also counting the partners of the children
    h += 'root'
    h += '.sum(d => (d.value || 0) + (d.partners || []).length)'
    h += '.sort((a, b) => b.value - a.value);' + "\n"

    h += 'const tree = d3.tree().size([height, width * 1.2])'
    h += '.separation((a, b) => {'
    #// More separation between nodes with many children
    h += 'const totalPartners = (a.data.partners || []).length + (b.data.partners || []).length;'
    h += 'return 1 + (totalPartners / 5);'
    h += '});' + "\n"

    h += 'draw(tree(root));' + "\n"

    h += 'function collapse(d) {'
    h += 'if (d.children) {'
    h += 'd._children = d.children;'
    h += 'd._children.forEach(collapse);'
    h += 'd.children = null;'
    h += '}'
    h += '}'
    h += 'function click(d) {'
    h += 'if (d.children) {'
    h += 'd._children = d.children;'
    h += 'd.children = null;'
    h += '} else {'
    h += 'd.children = d._children;'
    h += 'd._children = null;'
    h += '}'
    h += 'update(d);'
    h += '} '
    h += "</script>"
    h += '</div></div></div>'
    return h
